/******************
 * Misc utils.    *
 ******************/

function assert(cond, msg) {
    if (!cond) {
        log(msg);
        throw new Error();
    }
}

function timeit(label, func) {
    // Check whether profiling is available first.
    if (typeof debugPerfStart === "function") {
        let id = debugPerfStart(label);
        func();
        debugPerfEnd(id);
    } else {
        func();
    }
}

/******************
 * Configuration. *
 ******************/

const mapWidth = 72; // Width of the map, in tiles.
const mapHeight = 108; // Height of the map, in tiles.
const mapArea = mapWidth * mapHeight; // Total number of tiles on the map.
const borderSize = 4;
const baseAreaSize = 16;

const players = 2;
const scavengersPlayer = (players > 7 ? 10 : 7);

const Symmetry = {
    NONE: 'NONE',
};

const symmetry = Symmetry.NONE;

const waterLevel = (players == 2) ? 230 : 280; // Terrain below this level is water.
const grassLevel = (players == 2) ? 550 : 600;
const grassSnowLevel = (players == 2) ? 630 : 615; // Terrain above this level is unpassable.
const cliffLevel = (players == 2) ? 740 : 700;
const gravelSnowLevel = 850;
const maxLevel = 1000;

const grassHeight = 112;
const gravelMinHeight = 192;

const perlinStartSize = (players == 2) ? 12 : 16;
const perlinCrispness = 4;

// How many times to try to place a large object set (eg., player base)
// before throwing away the map and starting from scratch.
const placementRetries = 300;

// Try to place one oil and one village in each quadrant of this size.
const oilQuadrant = 24;
const minOils = 5;
const minScavOils = 2;
const minVillages = 5;
const minScavVillages = 1;

// Values over (512 - gravelMinHeight) are clipped. Some clipping is ok.
const mountainHeight = 550;

// Controls placement of dirt on grass.
const nonDirtiness = 20;

// Shortcuts for map tile textures.
const Tile = {
    GRASS: 0,
    WATER: 17,
    CONCRETE: 22,
    // This is a 25% gravel 75% water tile but with ttype set to cliff.
    // This gives a better texture (water bed) than 100% gravel.
    GRAVEL: 32,
    GRAVEL_SNOW: 41,
    DIRT: 53,
    DARK_DIRT: 5,
    GRASS_SNOW: 23,
    SNOW: 64,
    CONCRETE: 22,

    GRASS_SNOW_PATCH_1: 47,
    CRASS_SNOW_PATCH_2: 74,
    GRASS_CRATER: 58,

    DIRT_ROCKS_1: 27,
    DIRT_ROCKS_2: 28,
    DIRT_CRATER: 56,

    GRAVEL_GRASS_PATCH: 55,
    GRAVEL_DIRT_PATCH: 57,
    GRAVEL_SNOW_PATCH_1: 8,
    GRAVEL_SNOW_PATCH_2: 43,
    GRAVEL_CRATER: 62,

    GRAVEL_SNOW_SNOW_PATCH_1: 73,
    GRAVEL_SNOW_SNOW_PATCH_2: 75,

    SNOW_GRAVEL_PATCH: 70,

    ROAD: 59,
    ROAD_END: 60,
    ROAD_T: 13,
    ROAD_TRACKS: 37,

    TRACK_1: 49,
    TRACK_2: 51,
    TRACK_TURN: 50,
    TRACK_END: 52,
};

const baseAreas = [
    [borderSize, mapHeight - baseAreaSize - borderSize, baseAreaSize, mapHeight - 1 - borderSize],
    [mapWidth - baseAreaSize - borderSize, borderSize, mapWidth - 1 - borderSize, baseAreaSize + borderSize],
    [borderSize, borderSize, baseAreaSize + borderSize, baseAreaSize + borderSize],
    [mapWidth - baseAreaSize - borderSize, mapHeight - baseAreaSize - borderSize, mapWidth - 1 - borderSize, mapHeight - 1 - borderSize]
];

function inBaseArea(x, y) {
    for (let i = 0; i < players; ++i) {
        let area = baseAreas[i];
        if (x >= area[0] && y >= area[1] && x <= area[2] && y <= area[3])
            return true;
    }

    return false;
}


/*********************
 * Noise generation. *
 *********************/

// Ok, this is a fairly basic fractal value noise implementation. It splits the map
// into small quadrants and constructs a quadratic curve for every quadrant
// that interpolates the four entirely random values at every vertex surrounding
// this quadrant. The output noise consists of such interpolations added up over
// smaller and smaller quadrants.
class Noise {
    constructor(width, height) {
        this.width = width;
        this.height = height;

        this.data = Array(width * height).fill(0);

        // Use the native generator if available.
        if (typeof generateFractalValueNoise === "function") {
            let riggedRegions = [];

            // Borders are rigged to mountainous so that to make it
            // visually apparent that they're believably impassable.
            riggedRegions = riggedRegions.concat([
                [0, 0, width - 1, 0, (x, y, i, range) => range - 1],
                [0, 0, 0, height - 1, (x, y, i, range) => range - 1],
                [width - 1 - 0, 0, width - 1, height - 1, (x, y, i, range) => range - 1],
                [0, height - 1 - 0, width - 1, height - 1, (x, y, i, range) => range - 1]
            ]);

            // Rig player bases to be flat.
            for (let i = 0; i < players; ++i) {
                let area = baseAreas[i];
                riggedRegions.push([area[0], area[1], area[2], area[3],
                                    (x, y, i, range) => range * 4 / 10]);
            }

            this.data = generateFractalValueNoise(width, height,
                                                  /*range=*/1000,
                                                  perlinCrispness,
                                                  perlinStartSize,
                                                  /*normalizeToRange=*/1000,
                                                  riggedRegions);
            return;
        }

        // Generate manually if the native generator is not available.
        // This is much slower.
        let tw = perlinStartSize, th = perlinStartSize, range = 1024, idx = 0;
        while (tw >= 2 && th >= 2 && range >= 1) {
            tw = Math.floor(tw / 2);
            th = Math.floor(th / 2);

            timeit("noise iteration " + tw + "x" + th, () => {
                let ta = tw * th;
                let tw1 = tw - 1;
                let th1 = th - 1;

                range = Math.floor(range * perlinCrispness / 10);
                idx += 1;

                let lw = Math.floor(width / tw) + 1;
                let lh = Math.floor(height / th) + 1;
                let layer = Array(lw * lh);
                timeit("noise layer", () => {
                    for (let x = 0; x < lw; ++x) {
                        for (let y = 0; y < lh; ++y) {
                            if (x == 0 || y == 0 || x == lw - 1 || y == lh - 1) {
                                layer[x * lh + y] = range;
                            } else if (inBaseArea(x * tw, y * th)) {
                                // Make it easier to spawn player bases.
                                layer[x * lh + y] = Math.floor(range * 2 / 5);
                            } else {
                                layer[x * lh + y] = gameRand(range);
                            }
                        }
                    }
                });

                // Despite four nested loops, this is linear over map area
                // as it visits every cell exactly once.
                for (let ix = 0; ix < lw - 1; ++ix) {
                    for (let iy = 0; iy < lh - 1; ++iy) {
                        let tx = ix * tw;
                        let ty = iy * th;

                        let topLeft = layer[ix * lh + iy] / ta,
                            topRight = layer[(ix + 1) * lh + iy] / ta,
                            bottomLeft = layer[ix * lh + iy + 1] / ta,
                            bottomRight = layer[(ix + 1) * lh + iy + 1] / ta;

                        for (let x = 0; x < tw; ++x) {
                            for (let y = 0; y < th; ++y) {
                                this.data[this.height * (tx + x) + (ty + y)] += Math.floor(
                                    bottomRight * x * y +
                                    topRight * x * (th1 - y) +
                                    bottomLeft * (tw1 - x) * y +
                                    topLeft * (tw1 - x) * (th1 - y));
                            }
                        }
                    }
                }
            });
        }
    }

    normalize() {
        timeit("noise normalization", () => {
            let min = Infinity, max = 0;
            for (let x = 0; x < this.width; ++x) {
                for (let y = 0; y < this.height; ++y) {
                    let idx = this.height * x + y;
                    max = Math.max(max, this.data[idx]);
                    min = Math.min(min, this.data[idx]);
                }
            }
            for (let x = 0; x < this.width; ++x) {
                for (let y = 0; y < this.height; ++y) {
                    let idx = this.height * x + y;
                    this.data[idx] = Math.floor((this.data[idx] - min) * maxLevel / (max + 1));
                    assert(this.data[idx] >= 0 && this.data[idx] < maxLevel,
                           "Perlin noise not normalized");
                }
            }
        });
    }
}

/***********************
 * Village generation. *
 ***********************/

const Hut = {
    EMPTY: "EMPTY",
    HUT: "HUT",
    TRACK: "TRACK",
    ROAD: "ROAD"
};

class VillageData {
    constructor(radius) {
        this.radius = radius;

        let attempt = 1;
        while (!this.generate()) {
            ++attempt;
        }
    }

    generate() {
        this.data = [];

        for (let x = -this.radius - 2; x <= this.radius + 2; ++x) {
            this.data[x] = [];
            for (let y = -this.radius - 2; y <= this.radius + 2; ++y) {
                this.data[x][y] = Hut.EMPTY;
            }
        }

        // Start with a track. We're going to lay only tracks at first because
        // they have better connections: namely, they offer a turn tile
        // whereas roads don't. We will later replace some tracks with roads.
        this.data[0][0] = Hut.TRACK;

        // A single road tile is not a valid village. Immediately extend
        // that track road in a random direction.
        let initialDir = [[0, 1], [0, -1], [1, 0], [-1, 0]][gameRand(4)];
        let initialDirX = initialDir[0], initialDirY = initialDir[1];
        this.data[initialDirX][initialDirY] = Hut.TRACK;


        let queue = [[0, 0], initialDir];
        let numTracks = 2, hutList = [];

        while (queue.length > 0) {
            let idx = gameRand(queue.length);
            let item = queue[idx];
            let x = item[0], y = item[1];
            queue.splice(idx, 1);

            // Stop growing at radius.
            if (Math.abs(x) >= this.radius || Math.abs(y) >= this.radius)
                continue;

            this.iterNeighbors(x, y, (xx, yy) => {
                if (this.data[xx][yy] != Hut.EMPTY)
                    return;

                // Any track set is valid as long as it doesn't have any
                // 3-way or 4-way crossings. We're about to place a road here so
                // we shouldn't do it if it causes a crossing to appear.
                let maxNeighborNeighbors = this.countTrackNeighbors(xx, yy);
                this.iterNeighbors(xx, yy, (xxx, yyy) =>
                    maxNeighborNeighbors = Math.max(
                        maxNeighborNeighbors,
                        this.countTrackNeighbors(xxx, yyy)));
                if (maxNeighborNeighbors < 2 && gameRand(3) == 0) {
                    this.data[xx][yy] = Hut.TRACK;
                    queue.push([xx, yy]);
                    ++numTracks;
                } else if (gameRand(2) > 0) {
                    this.data[xx][yy] = Hut.HUT;
                    hutList.push([xx, yy]);
                }
            });
        }

        if (numTracks < this.radius / 2) {
            // Too few tracks generated. Retry.
            return false;
        }

        // TODO: Place roads.

        if (hutList.length < this.radius || hutList.length < 6) {
            // Too few huts generated. Retry.
            // We need at least 5 huts to fit scavengers in.
            return false;
        }

        // Scavenger factory is 1x2. Place it carefully.
        this.scavFactoryCandidate = null;
        for (let idx = 0; idx < hutList.length; ++idx) {
            let x = hutList[idx][0], y = hutList[idx][1];
            if (x - 1 >= -this.radius && this.data[x - 1][y] == Hut.EMPTY) {
                this.scavFactoryCandidate = [x, y];
                hutList.splice(idx, 1);
                break;
            }
        }

        if (this.scavFactoryCandidate === null) {
            // No scav factory. Retry. This is rare.
            return false;
        }

        let scavGeneratorIdx = gameRand(hutList.length);
        this.scavGeneratorCandidate = hutList[scavGeneratorIdx];
        hutList.splice(scavGeneratorIdx, 1);

        this.scavBunkerCandidates = [];
        for (let i = 0; i < 3; ++i) {
            let idx = gameRand(hutList.length);
            this.scavBunkerCandidates.push(hutList[idx]);
            hutList.splice(idx, 1);
        }

        return true;
    }

    acceptScavCandidates() {
        // Scavenger candidates have been read and taken care of.
        // Eliminate the respective huts.
        let accept = (xy, plusOne = false) => {
            this.data[xy[0]][xy[1]] = Hut.EMPTY;
            if (plusOne)
                this.data[xy[0] - 1][xy[1]] = Hut.EMPTY;
        }

        accept(this.scavFactoryCandidate, true);
        accept(this.scavGeneratorCandidate);
        for (xy of this.scavBunkerCandidates)
            accept(xy);
    }

    iterNeighbors(x, y, callback) {
        callback(x - 1, y);
        callback(x + 1, y);
        callback(x, y - 1);
        callback(x, y + 1);
    }

    countNeighbors(x, y, callback) {
        let count = 0;
        this.iterNeighbors(x, y, (xx, yy) => {
            if (callback(xx, yy))
                ++count;
        });
        return count;
    }

    countTrackNeighbors(x, y) {
        return this.countNeighbors(x, y, 
            (xx, yy) => (this.data[xx][yy] == Hut.TRACK));
    }
}


/*******************************************************
 * An auxiliary class to keep track of occupied tiles. *
 *******************************************************/

class OccupancyMap {
    constructor(width, height) {
        this.width = width;
        this.height = height;

        // Tracks whether each tile is occupied or not.
        this.data = new Array(width);

        // Tracks whether the given tile is on the same continent
        // as any specific tile passed to floodFillContinent().
        // This is used for verifying reachability from one player to another.
        // Initially uninitialized - use floodFillContinent() for the first time
        // to initialize.
        this.continent = new Array(width);

        // Precomputed distance from every unoccupied tile to the nearest
        // occupied tile. These are expensive to compute but make finding
        // a large unoccupied spot much faster.
        // Initially unset; in order to take advantage of them you need to
        // call precomputeDistances() without arguments first. Once initialized,
        // they are kept up to date every time a square gets occupied
        // via occupyRadius().
        this.distance = null;

        for (let x = 0; x < width; ++x) {
            this.continent[x] = new Array(height).fill(false);

            // Pre-occupy borders.
            if (x < borderSize || x >= width - borderSize) {
                this.data[x] = new Array(height).fill(true);
            } else {
                this.data[x] = new Array(height).fill(false);
                for (let y = 0; y < borderSize; ++y) {
                    this.data[x][y] = true;
                    this.data[x][height - 1 - y] = true;
                }
            }
        }
    }

    occupy(x, y) {
        // Note: this does not update the distances map or the continent map.
        // This can be dangerous or useful.
        this.data[x][y] = true;
    }

    vacate(x, y) {
        // Note: this does not update the distances map or the continent map.
        // This can be dangerous or useful.
        this.data[x][y] = false;
    }

    isOccupied(x, y) {
        return this.data[x][y];
    }

    isRadiusOccupied(centerX, centerY, r) {
        let x1 = Math.max(centerX - r, 0);
        let x2 = Math.min(centerX + r, this.width - 1);
        let y1 = Math.max(centerY - r, 0);
        let y2 = Math.min(centerY + r, this.height - 1);
        for (let x = x1; x <= x2; ++x) {
            for (let y = y1; y <= y2; ++y) {
                if (this.data[x][y])
                    return true;
            }
        }

        return false;
    }

    occupyRadius(centerX, centerY, r) {
        // Note: this does not update the continent map.
        // This can be dangerous or useful.
        let borderList = [];
        for (let x = centerX - r; x <= centerX + r; ++x) {
            if (x <= 0 || x >= this.width - 2)
                continue;

            for (let y = centerY - r; y <= centerY + r; ++y) {
                if (y <= 0 || y >= this.height - 2)
                    continue;

                this.data[x][y] = true;
                this.distance[x][y] = 0;
                if (Math.abs(x - centerX) == r || Math.abs(y - centerY) == r) {
                    borderList.push([x, y]);
                }
            }
        }
        if (this.distance !== null) {
            this.precomputeDistances(borderList);
        }
    }

    floodFillContinent(contX, contY, radius) {
        // This function does not rely on precomputed distances.
        // It could but it's cheaper to refute a map early
        // based on connectivity alone than to precompute distances
        // and that's the single most common reason to refute a map.
        let isValidTile = (x, y) => {
            if (x < radius || x >= this.width - 1 - radius ||
                y < radius || y >= this.height - 1 - radius) {
                return false;
            }
            for (let xx = x - radius; xx <= x + radius; ++xx)
                for (let yy = y - radius; yy <= y + radius; ++yy)
                    if (this.data[xx][yy])
                        return false;
            return true;
        }

        let queue = [[contX, contY]];
        while (queue.length > 0) {
            let item = queue.pop();
            let currX = item[0];
            let currY = item[1];
            this.continent[currX][currY] = true;

            let maybeEnqueue = (x, y) => {
                if (!this.continent[x][y] && isValidTile(x, y))
                    queue.push([x, y]);
            }

            maybeEnqueue(currX - 1, currY);
            maybeEnqueue(currX + 1, currY);
            maybeEnqueue(currX, currY - 1);
            maybeEnqueue(currX, currY + 1);
        }
    }

    precomputeDistances(queue = null) {
        if (queue === null) {
            this.distance = new Array(this.width);
            for (let x = 0; x < this.width; ++x)
                this.distance[x] = new Array(this.height).fill(Infinity);

            queue = [];
            for (let x = 0; x < this.width; ++x) {
                for (let y = 0; y < this.height; ++y) {
                    if (this.data[x][y]) {
                        this.distance[x][y] = 0;
                        if (x > 0 && y > 0 && x < this.width - 1 && y < this.height - 1)
                            queue.push([x, y]);
                    }
                }
            }
        } else {
            assert(this.distance !== null,
                   "Distances not computed globally before local recomputation");
        }

        let queueStart = 0;
        while (queue.length - queueStart> 0) {
            let item = queue[queueStart];
            ++queueStart;

            let x = item[0], y = item[1];

            for (let neighbor of [[x - 1, y - 1], [x, y - 1], [x + 1, y - 1],
                                  [x - 1, y],                 [x + 1, y],
                                  [x - 1, y + 1], [x, y + 1], [x + 1, y + 1]]) {
                let xx = neighbor[0];
                let yy = neighbor[1];

                if (this.data[xx][yy]) {
                    continue;
                }

                let newDistance = this.distance[x][y] + 1;
                if (newDistance < this.distance[xx][yy]) {
                    this.distance[xx][yy] = newDistance;
                    queue.push([xx, yy]);
                }
            }
        }
    }

    findSpot(r, x1, y1, x2, y2) {
        let bestX, bestY, bestDistance = 0;
        for (let x = x1; x <= x2; ++x) {
            for (let y = y1; y <= y2; ++y) {
                if (this.distance[x][y] > bestDistance) {
                    bestX = x;
                    bestY = y;
                    bestDistance = this.distance[x][y];
                }
            }
        }

        if (bestDistance >= r)
            return [bestX, bestY];

        return null;
    }
}


            
/*****************************
 * Noise to map translation. *
 *****************************/

// Basic data structures of the map.
let texture;
let heightmap;
let structures;
let droids;
let features;

function clearMap() {
    texture = Array(mapArea).fill(Tile.GRASS);
    heightmap = Array(mapArea).fill(0);
    structures = [];
    droids = [];
    features = [];
}

// Magic constants corresponding to the game engine's constants.
const TILE_XFLIP = 0x8000;
const TILE_YFLIP = 0x4000;
const TILE_ROT = 0x1000;

const DROID_ROT = 0x4000;

function setTexture(tex, x, y, rot=gameRand(4), xflip=0, yflip=0) {
    texture[mapWidth * y + x] = tex | (rot * TILE_ROT)
                                    | (xflip * TILE_XFLIP)
                                    | (yflip * TILE_YFLIP);
}

function getTexture(x, y) {
    return texture[mapWidth * y + x] & 0xff;
}

function setVertexHeight(h, x, y) {
    heightmap[mapWidth * y + x] = h;
}

function getVertexHeight(x, y) {
    return heightmap[mapWidth * y + x];
}

function setTileHeight(h, x, y) {
    heightmap[mapWidth * y + x] = h;
    if (x < mapWidth - 1)
        heightmap[mapWidth * y + (x + 1)] = h;
    if (y < mapHeight - 1)
        heightmap[mapWidth * (y + 1) + x] = h;
    if (x < mapWidth - 1 && y < mapHeight - 1)
        heightmap[mapWidth * (y + 1) + (x + 1)] = h;
}

function addStructure(name, x, y, player, modules = 0, rot = 0) {
    structures.push({
        name: name,
        position: [128 * x, 128 * y],
        direction: 0x4000 * rot,
        modules: modules,
        player: player
    });
}

function addDroid(name, x, y, player, rot = 0) {
    droids.push({
        name: name,
        position: [128 * x + 64, 128 * y + 64],
        direction: rot * DROID_ROT,
        player: player
    });
}

function addFeature(name, x, y, rot = gameRand(3)) {
    features.push({
        name: name,
        position: [128 * x, 128 * y],
        direction: rot * DROID_ROT
    });
}

function tryGenerateMap() {
    clearMap();

    let midBase = borderSize + Math.floor(baseAreaSize / 2);

    // The noise corresponds to heightmap. The heightmap is one unit wider
    // than the tile map because tiles are between height vertices.
    let noise;
    timeit("noise generation", () => {
        noise = new Noise(mapWidth + 1, mapHeight + 1);
    });
    let occupancy = new OccupancyMap(mapWidth, mapHeight);

    function getVertexLevel(x, y) {
        return noise.data[noise.height * x + y];
    }
    function getTileLevel(x, y) {
        return Math.max(noise.data[noise.height * x + y],
                        noise.data[noise.height * (x + 1) + y],
                        noise.data[noise.height * x + (y + 1)],
                        noise.data[noise.height * (x + 1) + (y + 1)]);
    }

    for (let x = 0; x < mapWidth; ++x) {
        for (let y = 0; y < mapHeight; ++y) {
            let tileLevel = getTileLevel(x, y);
            if (tileLevel < waterLevel) {
                occupancy.occupy(x, y); // Underwater.
            } else if (tileLevel >= grassSnowLevel) {
                occupancy.occupy(x, y); // Mountain.
            }
        }
    }

    timeit("continents", () => {
        occupancy.floodFillContinent(midBase, mapHeight - 1 - midBase, 1);
    });
    if (!occupancy.continent[mapWidth - 1 - midBase][midBase]) {
        log("Bases not connected (early check)");
        return false;
    }

    timeit("distances", () => {
        occupancy.precomputeDistances();
    });

    for (let x = 0; x < mapWidth; ++x) {
        for (let y = 0; y < mapHeight; ++y) {
            // Vertex height depends only on the corresponding noise data point.
            let vertexLevel = getVertexLevel(x, y);
            if (vertexLevel < waterLevel) {
                setVertexHeight(0, x, y);
            } else if (vertexLevel < grassSnowLevel) {
                setVertexHeight(grassHeight * (vertexLevel - waterLevel)
                                            / (grassSnowLevel - waterLevel), x, y);
            } else {
                let vertexHeight = gravelMinHeight + Math.floor(
                    mountainHeight * (vertexLevel - grassSnowLevel)
                                   / (maxLevel - grassSnowLevel));
                // Steadily clip the value to avoid overflows while preserving looks.
                if (vertexHeight > 480)
                    vertexHeight = 480 + (vertexHeight - 480) / 4;
                // Still too much? Hard clip.
                if (vertexHeight > 510)
                    vertexHeight = 510;
                setVertexHeight(vertexHeight, x, y);
            }

            // Texture depends on the four data points in its corners instead.
            let tileLevel = getTileLevel(x, y);
            if (tileLevel < waterLevel) {
                setTexture(Tile.WATER, x, y);
            } else if (tileLevel < grassSnowLevel) {
                let middleness = Math.abs(2 * tileLevel - waterLevel - grassSnowLevel);
                let textureRoll = gameRand(middleness);
                if (textureRoll < (grassSnowLevel - waterLevel) / nonDirtiness) {
                    switch (gameRand(300)) {
                        case 0:
                        case 1:
                        case 2:
                            setTexture(Tile.DIRT_ROCKS_1, x, y);
                            break;
                        case 3:
                        case 4:
                        case 5:
                            setTexture(Tile.DIRT_ROCKS_2, x, y);
                            break;
                        case 6:
                            setTexture(Tile.DIRT_CRATER, x, y);
                            break;
                        default:
                            setTexture(Tile.DIRT, x, y);
                            break;
                    }
                } else if (textureRoll < 2 * (grassSnowLevel - waterLevel) / nonDirtiness) {
                        setTexture(Tile.DARK_DIRT, x, y);
                } else {
                    switch (gameRand(300)) {
                        case 0:
                        case 1:
                        case 2:
                            setTexture(Tile.GRASS_SNOW_PATCH_1, x, y);
                            break;
                        case 3:
                        case 4:
                        case 5:
                            setTexture(Tile.GRASS_SNOW_PATCH_2, x, y);
                            break;
                        case 6:
                            setTexture(Tile.GRASS_CRATER, x, y);
                            break;
                        default:
                            if (tileLevel < grassLevel || gameRand(2)) {
                                setTexture(Tile.GRASS, x, y);
                            } else {
                                setTexture(Tile.GRASS_SNOW, x, y);
                            }
                            if (3 * gameRand(grassSnowLevel - tileLevel) >
                                        2 * (grassSnowLevel - waterLevel)) {
                                // Do not set occupancy yet.
                                // We have to place player bases first
                                // for better performance.
                                addFeature("Tree3", x, y);
                            }
                            break;
                    }
                }
            } else {
                if (4 * gameRand(maxLevel - tileLevel) >
                            3 * (maxLevel - grassSnowLevel)) {
                    addFeature("TreeSnow3", x, y);
                }

                if (tileLevel < cliffLevel) {
                    // Rotation adds variety to the gravel/bottom mixture
                    // (cf. comment for Tile.GRAVEL).
                    switch (gameRand(300)) {
                        case 0:
                        case 1:
                            setTexture(Tile.GRAVEL_GRASS_PATCH, x, y);
                            break;
                        case 2:
                        case 3:
                            setTexture(Tile.GRAVEL_DIRT_PATCH, x, y);
                            break;
                        case 4:
                        case 5:
                        case 6:
                            setTexture(Tile.GRAVEL_SNOW_PATCH_1, x, y);
                            break;
                        case 7:
                        case 8:
                        case 9:
                            setTexture(Tile.GRAVEL_SNOW_PATCH_2, x, y);
                            break;
                        case 10: // Disabled: ugly.
                            // setTexture(Tile.GRAVEL_CRATER, x, y);
                            // break;
                        default:
                            if (2 * gameRand(tileLevel - grassSnowLevel) >
                                    cliffLevel - grassSnowLevel) {
                                setTexture(Tile.GRAVEL_SNOW, x, y);
                            } else {
                                setTexture(Tile.GRAVEL, x, y);
                            }
                            break;
                    }
                } else if (tileLevel < gravelSnowLevel) {
                    switch (gameRand(100)) {
                        case 0:
                            setTexture(Tile.GRAVEL_SNOW_SNOW_PATCH_1, x, y);
                            break;
                        case 1:
                            setTexture(Tile.GRAVEL_SNOW_SNOW_PATCH_2, x, y);
                            break;
                        default:
                            if (2 * gameRand(tileLevel - cliffLevel) >
                                    gravelSnowLevel - cliffLevel) {
                                setTexture(Tile.SNOW, x, y);
                            } else {
                                setTexture(Tile.GRAVEL_SNOW, x, y);
                            }
                            break;
                    }
                } else {
                    switch (gameRand(100)) {
                        case 0:
                            setTexture(Tile.SNOW_GRAVEL_PATCH, x, y);
                            break;
                        default:
                            setTexture(Tile.SNOW, x, y);
                            break;
                    }
                    setTexture(Tile.SNOW, x, y);
                }
            }
        }
    }

    // Smoother water.
    let changesMade;
    do {
        changesMade = false;
        for (let x = 1; x < mapWidth - 1; ++x) {
            for (let y = 1; y < mapHeight - 1; ++y) {
                if (getTexture(x, y) == Tile.WATER) {
                    let top = getTexture(x, y - 1);
                    let bottom = getTexture(x, y + 1);
                    if (top != Tile.WATER && bottom != Tile.WATER) {
                        setTexture(Tile.GRASS, x, y);
                        occupancy.vacate(x, y);
                        changesMade = true;
                    }
                    let left = getTexture(x - 1, y);
                    let right = getTexture(x + 1, y);
                    if (left != Tile.WATER && right != Tile.WATER) {
                        setTexture(Tile.GRASS, x, y);
                        occupancy.vacate(x, y);
                        changesMade = true;
                    }
                }
            }
        }
    } while (changesMade);

    // Ok, now is the fun part. Place player bases.
    function placePlayerBase(player, x1, y1, x2, y2) {
        let spot = occupancy.findSpot(1, x1, y1, x2, y2);
        if (spot === null) {
            log("Unable to find spot for player " + player);
            return false;
        }

        let baseX = spot[0], baseY = spot[1];

        // Check that players are reachable from each other.
        if (!occupancy.continent[baseX][baseY]) {
            log("Bases not connected");
            return false;
        }

        occupancy.occupyRadius(baseX, baseY, 13);
        for (let x = baseX - 6; x <= baseX + 6; ++x) {
            for (let y = baseY - 7; y <= baseY + 6; ++y) {
                if (gameRand(3) == 0)
                    setTexture(Tile.CONCRETE, x, y)
                else
                    setTexture(Tile.DIRT, x, y);
                setTileHeight(Math.floor(grassHeight / 2), x, y);
            }
        }

        addDroid("ConstructionDroid", baseX - 1, baseY + 1, player);
        addDroid("ConstructionDroid", baseX, baseY + 1, player);

        function addBaseStructure(str, x, y, kind = 0, modules = 0, rot = 0) {
            if (player == 0 || (players > 2 && player == 1))
                addStructure(str, baseX + x, baseY + y, player, modules, rot);
            else
                addStructure(str, baseX - x + (kind == 2), baseY - y + (kind >= 1), player, modules, rot);
        }

        addBaseStructure("A0CommandCentre", 0, 0, 2);
        addBaseStructure("A0LightFactory", 3, 0, 0, 2);
        addBaseStructure("A0LightFactory", 3, -4, 0, 2);
        addBaseStructure("A0CyborgFactory", 4, 4, 1);
        addBaseStructure("A0CyborgFactory", 2, 4, 1);
        addBaseStructure("A0PowerGenerator", 0, -3, 2, 1);
        addBaseStructure("A0ResearchFacility", -3, -4, 2, 1);
        addBaseStructure("A0ResearchFacility", -3, -1, 2, 1);
        addBaseStructure("A0ResearchFacility", -3, 2, 2, 1);

        addBaseStructure("A0ResourceExtractor", 0, 4);
        addBaseStructure("A0ResourceExtractor", -2, 4);
        addBaseStructure("A0ResourceExtractor", -4, 4);
        addBaseStructure("A0ResourceExtractor", -1, 2);

        addBaseStructure("WallTower02", -6, -7);
        addBaseStructure("WallTower02", 6, -7);
        addBaseStructure("WallTower02", -6, 6);
        addBaseStructure("WallTower02", 6, 6);

        addBaseStructure("A0HardcreteMk1Wall", -5, -7);
        addBaseStructure("A0HardcreteMk1Wall", -6, -6, 0, 0, 1);
        addBaseStructure("A0HardcreteMk1Wall", 5, -7);
        addBaseStructure("A0HardcreteMk1Wall", 6, -6, 0, 0, 1);
        addBaseStructure("A0HardcreteMk1Wall", -5, 6);
        addBaseStructure("A0HardcreteMk1Wall", -6, 5, 0, 0, 1);
        addBaseStructure("A0HardcreteMk1Wall", 5, 6);
        addBaseStructure("A0HardcreteMk1Wall", 6, 5, 0, 0, 1);

        addBaseStructure("WallTower01", -4, -7);
        addBaseStructure("WallTower01", -6, -5);
        addBaseStructure("WallTower01", 4, -7);
        addBaseStructure("WallTower01", 6, -5);
        addBaseStructure("WallTower01", -4, 6);
        addBaseStructure("WallTower01", -6, 4);
        addBaseStructure("WallTower01", 4, 6);
        addBaseStructure("WallTower01", 6, 4);
        return true;
    }

    let margin = 5;

    if (players == 2) {
        if (!placePlayerBase(0, borderSize + margin,
                                mapHeight - 1 - borderSize - baseAreaSize + margin,
                                borderSize + baseAreaSize - margin,
                                mapHeight - 1 - borderSize - margin)) {
            return false;
        }

        if (!placePlayerBase(1, mapWidth - 1 - borderSize - baseAreaSize + margin,
                                borderSize + margin,
                                mapWidth - 1 - borderSize - margin,
                                borderSize + baseAreaSize - margin)) {
            return false;
        }
    } else {
        if (!placePlayerBase(0, borderSize + margin,
                                mapHeight - 1 - borderSize - baseAreaSize + margin,
                                borderSize + baseAreaSize - margin,
                                mapHeight - 1 - borderSize - margin)) {
            return false;
        }
        if (!placePlayerBase(1, mapWidth - 1 - borderSize - baseAreaSize + margin,
                                mapHeight - 1 - borderSize - baseAreaSize + margin,
                                mapWidth - 1 - borderSize - margin,
                                mapHeight - 1 - borderSize - margin)) {
            return false;
        }
        if (!placePlayerBase(2, mapWidth - 1 - borderSize - baseAreaSize + margin,
                                borderSize + margin,
                                mapWidth - 1 - borderSize - margin,
                                borderSize + baseAreaSize - margin)) {
            return false;
        }
        if (!placePlayerBase(3, borderSize + margin, borderSize + margin,
                                borderSize + baseAreaSize - margin,
                                borderSize + baseAreaSize - margin)) {
            return false;
        }
    }

    // Place oils and villages. Quadrants start at the bottom left corner because
    // that's where the player base is so it's more important to be uniform.
    let oilsPlaced = 0;
    let scavOilsPlaced = 0;
    let villagesPlaced = 0;
    let scavVillagesPlaced = 0;
    for (let x = 0; x <= mapWidth - oilQuadrant; x += oilQuadrant) {
        for (let y = 0; y <= mapHeight - oilQuadrant; y += oilQuadrant) {
            let spot = occupancy.findSpot(1, x + 2, y + 2,
                                          x + oilQuadrant - 2, y + oilQuadrant - 2);
            if (spot === null)
                continue;

            let oilX = spot[0], oilY = spot[1];
            if (!occupancy.continent[oilX][oilY])
                continue;

            ++oilsPlaced;
            occupancy.occupyRadius(oilX, oilY, 1);

            if (gameRand(3) > 0) {
                addFeature("OilResource", oilX + 0.5, oilY + 0.5);
            } else {
                ++scavOilsPlaced;

                function addOilStructure(str, x, y, rot=0) {
                    addStructure(str, oilX + x, oilY + y,
                                 scavengersPlayer, /*modules=*/0, rot);
                }

                addOilStructure("A0ResourceExtractor", 0, 0);
                addOilStructure("A0BaBaGunTower", -1, -1, 1);
                addOilStructure("A0BaBaGunTower", 1, -1, 0);
                addOilStructure("A0BaBaGunTower", -1, 1, 2);
                addOilStructure("A0BaBaGunTower", 1, 1, 3);
                addOilStructure("A0BaBaHorizontalWall", -1, 0, 1);
                addOilStructure("A0BaBaHorizontalWall", 1, 0, 1);
                addOilStructure("A0BaBaHorizontalWall", 0, -1);
                addOilStructure("A0BaBaHorizontalWall", 0, 1);
            }

            // Generate villages similarly. Don't bother counting.
            let villageSize = 2 + gameRand(4);
            let villageSpot = occupancy.findSpot(villageSize + 1, x + 2, y + 2,
                                                 x + oilQuadrant - 2, y + oilQuadrant - 2);
            if (villageSpot === null)
                continue;

            let villageX = villageSpot[0], villageY = villageSpot[1];
            ++villagesPlaced;
            occupancy.occupyRadius(villageX, villageY, villageSize);

            function addVillageFeature(feat, x, y) {
                addFeature(feat, villageX + x + 0.5, villageY + y + 0.5, gameRand(4));
            }

            function setVillageTexture(tile, x, y, rot = 0) {
                setTexture(tile, villageX + x, villageY + y, rot);
            }

            function addVillageScavStructure(str, xy) {
                addStructure(str, villageX + xy[0], villageY + xy[1], scavengersPlayer);
                setTexture(Tile.CONCRETE, villageX + xy[0], villageY + xy[0]);
            }

            let village = new VillageData(villageSize);
            if (gameRand(3) == 0) {
                ++scavVillagesPlaced;
                addVillageScavStructure("A0BaBaFactory",
                                        village.scavFactoryCandidate);
                addVillageScavStructure("A0BaBaPowerGenerator",
                                        village.scavGeneratorCandidate);
                for (xy of village.scavBunkerCandidates)
                    addVillageScavStructure("A0BaBaBunker", xy);

                village.acceptScavCandidates();
            }

            for (let vx = -villageSize; vx <= villageSize; ++vx) {
                for (let vy = -villageSize; vy <= villageSize; ++vy) {
                    switch (village.data[vx][vy]) {
                        case Hut.NONE:
                            break;
                        case Hut.HUT:
                            setVillageTexture(Tile.CONCRETE, vx, vy);
                            const huts = [
                                "LogCabin1",
                                "LogCabin2",
                                "LogCabin3",
                                "LogCabin4",
                                "LogCabin5",
                                "WaterTower",
                            ];
                            addVillageFeature(huts[gameRand(huts.length)], vx, vy);
                            break;
                        case Hut.TRACK:
                            let topTrack = (village.data[vx][vy - 1] == Hut.TRACK),
                                bottomTrack = (village.data[vx][vy + 1] == Hut.TRACK),
                                leftTrack = (village.data[vx - 1][vy] == Hut.TRACK),
                                rightTrack = (village.data[vx + 1][vy] == Hut.TRACK);
                            let numTracks = topTrack + bottomTrack
                                          + leftTrack + rightTrack;
                            assert(numTracks == village.countTrackNeighbors(vx, vy),
                                   "Incorrect track neighbor calculation");
                            switch (numTracks) {
                                case 1:
                                    if (topTrack) {
                                        setVillageTexture(Tile.TRACK_END, vx, vy, 2);
                                    } else if (bottomTrack) {
                                        setVillageTexture(Tile.TRACK_END, vx, vy, 0);
                                    } else if (leftTrack) {
                                        setVillageTexture(Tile.TRACK_END, vx, vy, 1);
                                    } else if (rightTrack) {
                                        setVillageTexture(Tile.TRACK_END, vx, vy, 3);
                                    }
                                    break;
                                case 2:
                                    if (topTrack && bottomTrack) {
                                        setVillageTexture(
                                            gameRand(2) ? Tile.TRACK_1 : Tile.TRACK_2,
                                            vx, vy, gameRand(2) * 2);
                                    } else if (leftTrack && rightTrack) {
                                        setVillageTexture(
                                            gameRand(2) ? Tile.TRACK_1 : Tile.TRACK_2,
                                            vx, vy, gameRand(2) * 2 + 1);
                                    } else if (topTrack && leftTrack) {
                                        setVillageTexture(Tile.TRACK_TURN, vx, vy, 2);
                                    } else if (topTrack && rightTrack) {
                                        setVillageTexture(Tile.TRACK_TURN, vx, vy, 3);
                                    } else if (bottomTrack && leftTrack) {
                                        setVillageTexture(Tile.TRACK_TURN, vx, vy, 1);
                                    } else if (bottomTrack && rightTrack) {
                                        setVillageTexture(Tile.TRACK_TURN, vx, vy, 0);
                                    }
                                    break;
                                default:
                                    assert(false, "Invalid track layout: track has "
                                                  + numTracks + " neighbors");
                            }
                            break;
                    }
                }
            }
        }
    }
    if (oilsPlaced < minOils) {
        log("Too few oils placed (" + oilsPlaced + ")");
        return false;
    }
    if (scavOilsPlaced < minScavOils) {
        log("Too few scavenger oils placed (" + scavOilsPlaced + ")");
        return false;
    }
    if (villagesPlaced < minVillages) {
        log("Too few villages placed (" + villagesPlaced + ")");
        return false;
    }
    if (scavVillagesPlaced < minScavVillages) {
        log("Too few scavenger villages placed (" + scavVillagesPlaced + ")");
        return false;
    }

    // Flatten tiles under trees and set occupancy.
    for (let i = 0; i < features.length;) {
        if (features[i].name != "Tree3") {
            ++i;
            continue;
        }

        let x = Math.floor(features[i].position[0] / 128);
        let y = Math.floor(features[i].position[1] / 128);

        if (occupancy.isOccupied(x, y)) {
            features.splice(i, 1);
            continue;
        }

        ++i;
        occupancy.occupy(x, y);

        if (x >= mapWidth - 1 || y >= mapHeight - 1)
            continue;

        let h = Math.max(getVertexHeight(x, y),
                         getVertexHeight(x + 1, y),
                         getVertexHeight(x, y + 1),
                         getVertexHeight(x + 1, y + 1));
        setTileHeight(h, x, y);
    }


    // Return the data.
    setMapData(mapWidth, mapHeight, texture, heightmap, structures, droids, features);

    return true;
}

timeit("total time", () => {
    let attempt = 1;
    while (true) {
        let result;
        timeit("attempt " + attempt + " time", () => {
            result = tryGenerateMap();
        });
        if (result)
            break;
        ++attempt;
        log("Attempt " + attempt);
    }
});
